/* Copyright 1995 Jim McBeath
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that my name not be used in advertising
 * or publicity pertaining to distribution of the software without
 * specific, written prior permission.  I make no representations about
 * the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * I disclaim all warranties with regard to this software, including all
 * implied warranties of merchantability and fitness.  In no event shall
 * I be liable for any special, indirect or consequential damages or any
 * damages whatsoever resulting from loss of use, data or profits, whether
 * in an action of contract, negligence or other tortious action, arising
 * out of or in connection with the use or performance of this software.
 */
/* urlgoph.c - gopher access for url package
 *
 * Jim McBeath, September 18, 1995
 */

#include <varargs.h>
#include "urlopen.h"

#define crlf "\r\n"

#define URL_GOPHER_TEXT       '0'
#define URL_GOPHER_DIRECTORY  '1'
#define URL_GOPHER_CSO        '2'
#define URL_GOPHER_ERROR      '3'
#define URL_GOPHER_MACBINHEX  '4'
#define URL_GOPHER_DOSBIN     '5'
#define URL_GOPHER_UUENCODED  '6'
#define URL_GOPHER_INDEX      '7'
#define URL_GOPHER_TELNET     '8'
#define URL_GOPHER_BINARY     '9'
#define URL_GOPHER_DUPLICATE  '+'
#define URL_GOPHER_IMAGE      'I'
#define URL_GOPHER_MIME       'M'
#define URL_GOPHER_TN3270     'T'
#define URL_GOPHER_GIF        'g'
#define URL_GOPHER_HTML       'h'
#define URL_GOPHER_INLINE     'i'
#define URL_GOPHER_SOUND      's'
#define URL_GOPHER_WWW        'w'
#define URL_GOPHER_PLUS_IMAGE ':'
#define URL_GOPHER_PLUS_MOVIE ';'
#define URL_GOPHER_PLUS_SOUND '<'
#define URL_GOPHER_EOF        '.'

extern char *getenv();

/* forward decls */
int UrlGopherOpen();
char *UrlGopherGetLine();
char *UrlGopherGetMeta();
int UrlGopherIsFiltered();
static int UrlGopherSendGet();

typedef struct _gopherinfo {
	int type;	/* gopher type character */
	int isdir;	/* true if type==URL_GOPHER_DIRECTORY */
	int state;	/* for menu parsing */
} GopherInfo;

void
UrlGopherRegister(ctx)
UrlContext *ctx;
{
	UACCESS *access;

	access = UrlRegisterNew(ctx,"gopher");
	if (!access) return;
	access->datasize = sizeof(GopherInfo);
	access->open = UrlGopherOpen;
	access->getline = UrlGopherGetLine;
	access->getmeta = UrlGopherGetMeta;
	access->isfiltered = UrlGopherIsFiltered;
	access->default_port = "70";
}

int	/* status */
UrlGopherOpen(uf)
UFILE *uf;
{
	int sc;
	int t;
	GopherInfo *td;	/* our access-specific data */

	td = (GopherInfo *)(uf->tdata);
	if (!uf->parse->port)
		uf->parse->port = "70";	/* default port as string */
	sc = UrlConnect(uf);		/* make the connection */
	if (sc<0)
		return UrlGetErrno(uf->ctx);
	uf->fd = sc;
	t = UrlGopherSendGet(uf);
	if (t<0) {
		close(sc);
		uf->fd = -1;
		return UrlGetErrno(uf->ctx);
	}
	/* if header, read it here */
	return 0;
}

static
int	/* status */
UrlGopherSendGet(uf)
UFILE *uf;
{
	int type;
	static char *gbuf=0;
	static int gbuflen=0;
	char *p;
	int l;
	char *file;
	char *query;
	int t;
	GopherInfo *td;	/* our access-specific data */

	td = (GopherInfo *)(uf->tdata);
	/* set up our buffer */
	l = strlen(uf->parse->url)+40;
	if (l>gbuflen) {
		if (gbuf)
			UrlFree(gbuf);
		gbuf = (char *)UrlMalloc(l);
		if (!gbuf)
			return UrlSetErrno(uf->ctx,ENOMEM);
		gbuflen = l;
	}
	p = gbuf;

	type = '1';	/* default type */
	file = uf->parse->file;
	query = uf->parse->query;
	if (!query) query="";
	if (!file) file="/";
	if (*file=='/') {
		if (*++file)
			type = *file++;
	}
	if (type==URL_GOPHER_INDEX) {
		strcpy(p,file);
		p += strlen(p);
		*p++ = '\t';
		p += strlen(p);
		strcpy(p,query);
		p += strlen(p);
	} else if (type==URL_GOPHER_CSO) {
		strcpy(p,"query ");
		p += strlen(p);
		strcpy(p,query);
		p += strlen(p);
	} else {
		strcpy(p,file);
		p += strlen(p);
	}
	/* TBD - need to un-escape the command */
	/* TBD - don't allow CR or LF in string */

	strcpy(p,crlf);		/* blank line means end of header */
	p += strlen(p);
	td->type = type;	/* save gopher type for later decoding */
	td->isdir = (type==URL_GOPHER_DIRECTORY);
	UrlMessage(uf->ctx,URL_MSG_CONNECT,"Sending Gopher request");
	if (uf->ctx->debug)
		UrlMessagef(uf->ctx,URL_MSG_DEBUG,"Send to gopher: %s",gbuf);
	t = write(uf->fd,gbuf,p-gbuf);
	if (t<0)
		return UrlSetErrno(uf->ctx,errno);
	return 0;	/* header successfully sent */
}

char *
UrlGopherGetMeta(uf,name)
UFILE *uf;
char *name;
{
	UrlContext *ctx;
	GopherInfo *td;
	char *ct;

	td = (GopherInfo *)(uf->tdata);
	ctx = uf->ctx;
	if (strcmp(name,"content-type")==0) {
		switch (td->type) {
		case URL_GOPHER_TEXT:
			ct = "text/plain"; break;
		case URL_GOPHER_HTML:
		case URL_GOPHER_DIRECTORY:
		case URL_GOPHER_CSO:
		case URL_GOPHER_INDEX:
			ct = "text/html"; break;
		case URL_GOPHER_MACBINHEX:
		case URL_GOPHER_DOSBIN:
		case URL_GOPHER_UUENCODED:
		case URL_GOPHER_BINARY:
			ct = "application/binary"; break;
		case URL_GOPHER_TELNET:
		case URL_GOPHER_TN3270:
			ct = "application/telnet"; break;
		case URL_GOPHER_IMAGE:
		case URL_GOPHER_GIF:
		case URL_GOPHER_PLUS_IMAGE:
			ct = "image/gif"; break;
		case URL_GOPHER_PLUS_MOVIE:
			ct = "video/unknown"; break;
		case URL_GOPHER_SOUND:
		case URL_GOPHER_PLUS_SOUND:
			ct = "audio/basic"; break;
		case URL_GOPHER_DUPLICATE:
		case URL_GOPHER_INLINE:
		case URL_GOPHER_MIME:
		case URL_GOPHER_WWW:
		case URL_GOPHER_EOF:
		case URL_GOPHER_ERROR:
		default:
			ct = (char *)0; break;
		}
		return ct;
	}
	UrlSetErrno(uf->ctx,EINVAL);
	return (char *)0;
}

int
UrlGopherIsFiltered(uf)
UFILE *uf;
{
	GopherInfo *td;

	td = (GopherInfo *)(uf->tdata);
	return (td->isdir);
}

int			/* number of fields found */
UrlGopherSplitFields(s,pp,np)
char *s;		/* the string to split - puts nulls into it in place */
char **pp;		/* array to fill with pointers */
int np;			/* size of pp array */
{
	return UrlSplitFields(s,'\t',0,pp,np);
}

/* Parse the line coming back from the menu listing, return HTML */
char *
UrlGopherParseMenuLine(ctx,rawline)
UrlContext *ctx;
char *rawline;
{
#define GOPHERMAXDIRFIELDS 20
	char *fields[GOPHERMAXDIRFIELDS];
	int n;
	int type;
#define GOPHERHTMLBUFSIZE 1000
	static char buf[GOPHERHTMLBUFSIZE];
	char portstr[100];
	char *cont;
	char *icon;
	char *icon_prefix;
	char *sep;

	if (!rawline || rawline[0]==0 || rawline[0]==URL_GOPHER_EOF)
		return (char *)0;
	type = rawline[0];
	n = UrlGopherSplitFields(rawline+1,fields,GOPHERMAXDIRFIELDS);
	if (n<4)
		return (char *)0;	/* bad format */
/* Gopher menu format:
  fields[0] = name
	[1] = selector
	[2] = host
	[3] = port
	[4] = extra junk
 */
	switch (type) {
	case URL_GOPHER_TEXT:
		cont = "gopher/plain"; break;
	case URL_GOPHER_DIRECTORY:
		cont = "gopher/menu"; break;
	case URL_GOPHER_ERROR:
		cont = "error"; break;
	case URL_GOPHER_CSO:
	case URL_GOPHER_INDEX:
		cont = "gopher/index"; break;
	case URL_GOPHER_MACBINHEX:
	case URL_GOPHER_DOSBIN:
	case URL_GOPHER_UUENCODED:
	case URL_GOPHER_BINARY:
		cont = "gopher/binary"; break;
	case URL_GOPHER_TELNET:
	case URL_GOPHER_TN3270:
		cont = "gopher/telnet"; break;
	case URL_GOPHER_IMAGE:
	case URL_GOPHER_GIF:
	case URL_GOPHER_PLUS_IMAGE:
		cont = "gopher/image"; break;
	case URL_GOPHER_PLUS_MOVIE:
		cont = "gopher/movie"; break;
	case URL_GOPHER_SOUND:
	case URL_GOPHER_PLUS_SOUND:
		cont = "gopher/sound"; break;
	case URL_GOPHER_DUPLICATE:
	case URL_GOPHER_HTML:
	case URL_GOPHER_INLINE:
	case URL_GOPHER_MIME:
	case URL_GOPHER_WWW:
	case URL_GOPHER_EOF:
	default:
		cont = "unknown"; break;
	}
	icon = UrlGetContIconType(ctx,cont);
	if (!icon)
		icon = "unknown";
	icon_prefix = ctx->icon_prefix;
	if (!icon_prefix)
		icon_prefix = "internal:icons/";
	switch (type) {
	case URL_GOPHER_WWW:
		sprintf(buf,
"<dt><img src=\"%s%s\">  <a href=\"%s\">%s</a>\n",
		    icon_prefix, icon, fields[1], fields[0]);
		return buf;
	case URL_GOPHER_TELNET:
		sep = (fields[1][0])?"@":"";
		sprintf(buf,
"<dt><img src=\"%s%s\">  <a href=\"telnet://%s%s%s\">%s</a>\n",
		    icon_prefix, icon, fields[1], sep, fields[2], fields[0]);
		return buf;
	case URL_GOPHER_TN3270:
		sep = (fields[1][0])?"@":"";
		sprintf(buf,
"<dt><img src=\"%s%s\">  <a href=\"tn3270://%s%s%s\">%s</a>\n",
		    icon_prefix, icon, fields[1], sep, fields[2], fields[0]);
		return buf;
	case URL_GOPHER_INLINE:
		sprintf(buf,"</dl><h2>%s</h2><dl>\n",fields[0]);
		return buf;
	default:
		if (fields[3][0]&&strcmp(fields[3],"0")!=0&&
		    strcmp(fields[3],"70")!=0) {
			sprintf(portstr,":%s",fields[3]);
		} else
			portstr[0] = 0;
		sprintf(buf,
"<dt><img src=\"%s%s\">  <a href=\"//%s%s/%c%s\">%s</a>\n",
			icon_prefix, icon,
			fields[2], portstr, type, fields[1], fields[0]);
		return buf;
	}
}

/* Parse a file and return HTML.
 * Separated from UrlGopherGetLine so that url_cat can open a local file
 * and pass it through this filter. */
char *
UrlGopherFGetHtmlLine(uf,statep)
UFILE *uf;
int *statep;
{
	char *s;
	UrlDirFileInfo fileinfo;
	static char buf[GOPHERHTMLBUFSIZE];

	if (uf->fd<0)
		return (char *)0;
	/* We're scanning a directory, so we need to parse the line
	 * and turn it into HTML */
	switch (*statep) {
	case 0:	/* need to read dir header, return HTML header */
		/* TBD - build a nicer title than this */
		*statep = 2;	/* continue header on next call */
		s = UrlUnparseC(uf->parse);
		sprintf(buf,
		   "<HTML><HEADER><TITLE>Gopher Menu for %s</TITLE></HEADER>\n",
			s);
		return buf;
	case 2:
		(*statep)++;
		s = UrlUnparseC(uf->parse);
		sprintf(buf,"<BODY><H1>Gopher Menu for %s</H1>\n",s);
		return buf;
	case 3:
		*statep = 1;	/* next time, start on menu items */
		return "<DL>\n";
	case 1: /* read menu line, return HTML for it */
		s = UrlUGetLineNoNl(uf);
		if (uf->ctx->debug)
			UrlMessagef(uf->ctx,URL_MSG_DEBUG,
				"Gopher receive line: %s",s?s:"(null)");
		s = UrlGopherParseMenuLine(uf->ctx,s);
		if (s)
			return s;
		*statep = -1;	/* done */
		/* finish it up */
		return "</DL></BODY></HTML>\n";
		break;
	default:
		errno = 0;	/* end of file */
		return (char *)0;	/* no more */
	}
}

char *
UrlGopherGetLine(uf)
UFILE *uf;
{
	char *rawline;
	GopherInfo *td;

	td = (GopherInfo *)(uf->tdata);
	if (!td->isdir) {
		rawline = UrlUGetLine(uf);
		if (uf->ctx->debug)
			UrlMessagef(uf->ctx,URL_MSG_DEBUG,
				"Gopher get raw line: %s",rawline);
		return rawline;
	}
	return UrlGopherFGetHtmlLine(uf,&(td->state));
}

/* end */
