/* Copyright 1995 Jim McBeath
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that my name not be used in advertising
 * or publicity pertaining to distribution of the software without
 * specific, written prior permission.  I make no representations about
 * the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * I disclaim all warranties with regard to this software, including all
 * implied warranties of merchantability and fitness.  In no event shall
 * I be liable for any special, indirect or consequential damages or any
 * damages whatsoever resulting from loss of use, data or profits, whether
 * in an action of contract, negligence or other tortious action, arising
 * out of or in connection with the use or performance of this software.
 */
/* urlhttp.c - http access for url package
 *
 * Jim McBeath, October 16, 1994
 */

#include "urlopen.h"
#include <ctype.h>

extern char *strchr();
extern char *UrlFGetLineNoNl();

#define crlf "\r\n"
#define HTTP_VERSION "HTTP/1.0"

typedef struct _httpinfo {
	UrlSymTab *st;	/* header meta-info */
} HttpInfo;

/* forward decls */
int UrlHttpOpen();
int UrlHttpClose();
static int UrlHttpSendGet();
char *UrlHttpGetMeta();

void
UrlHttpRegister(ctx)
UrlContext *ctx;
{
	UACCESS *access;

	access = UrlRegisterNew(ctx,"http");
	if (!access) return;
	access->datasize = sizeof(HttpInfo);
	access->open = UrlHttpOpen;
	access->close = UrlHttpClose;
	access->getmeta = UrlHttpGetMeta;
	access->default_port = "80";
}

/* Read the header lines from the http server */
static
int
UrlHttpReadHeader(uf)
UFILE *uf;
{
	UrlContext *ctx;
	UrlSymTab *st;
	HttpInfo *td;
	char *l, *c, *p;
	char *s;
	int hlen=0;

	ctx = uf->ctx;
	td = (HttpInfo *)(uf->tdata);
	st = UrlSymTabInit(ctx,uf->parse->url);
	if (!st) {
		UrlHttpClose(uf);
		return UrlGetErrno(uf->ctx);
	}
	UrlMessage(ctx,URL_MSG_CONNECT,"Waiting for HTTP response");
	td->st = st;
	l = UrlUGetLineNoNl(uf);	/* first line is version */
	if (!l) l="";
	if (ctx->debug)
		UrlMessagef(ctx,URL_MSG_DEBUG,"Received from HTTP: %s",l);
	UrlSymSet(ctx,st,"firstline",l);
/* TBD - need to check this line for status and proper syntax;
 * if not right, then we're probably talking to a pre-v1 server
 * (are there any still out there?) and there is no header being
 * returned to us. */
	hlen += strlen(l)+uf->xccount;
	while (!UrlUeof(uf)) {	/* or until we've got the whole header */
		l = UrlUGetLineNoNl(uf);
		if (!l || !*l)	/* blank line at end of header */
			break;
		if (ctx->debug)
		    UrlMessagef(ctx,URL_MSG_DEBUG,"Received from HTTP: %s",l);
		hlen += strlen(l)+uf->xccount;
		c = strchr(l,':');
		if (c) {
			*c++ = 0;
			if (isspace(*c))
				c++;
		} else
			c = "";
		for (p=l; *p; p++)
			if (isupper(*p))
				*p = tolower(*p);
		UrlSymSet(ctx,st,l,c);
	}
	hlen += uf->xccount;		/* blank line at end of header */
	UrlMessage(ctx,URL_MSG_CONNECT,"HTTP header received");
	s = UrlGetMeta(uf,"content-length");
	if (s) {
		int n;
		while (*s==' ') s++;
		n = atoi(s);
		if (n) {
			uf->filesize = n+hlen;	/* add header length */
		}
	}
	return 0;
}

int
UrlHttpOpen(uf)
UFILE *uf;
{
	int sc;
	int t;
	HttpInfo *td;	/* our access-specific data */
	UrlContext *ctx;

	ctx = uf->ctx;
	td = (HttpInfo *)(uf->tdata);
	if (ctx->use_proxy) {
		int pnum;
		if (ctx->use_proxy->port)
			pnum = atoi(ctx->use_proxy->port);
		else
			pnum = 80;
		sc = UrlConnectA(ctx,"http-proxy",ctx->use_proxy->host,pnum);
	} else {
		if (!uf->parse->port)
			uf->parse->port = "80";	/* default port as string */
		sc = UrlConnect(uf);		/* make the connection */
	}
	if (sc<0)
		return UrlGetErrno(ctx);
	uf->fd = sc;
	if (ctx->is_proxy)
		return 0;	/* let proxy client do raw send/receive */
	t = UrlHttpSendCmd(uf,"GET",HTTP_VERSION,1);
	if (t<0) {
		close(sc);
		uf->fd = -1;
		return UrlGetErrno(ctx);
	}
	t = UrlHttpReadHeader(uf);
	return t;
}

int
UrlHttpSendCmd(uf,cmd,postfix,addheader)
UFILE *uf;
char *cmd;
char *postfix;
int addheader;	/* true means add other header lines plus blank line */
{
	static char *gbuf=0;
	static int gbuflen=0;
	char *p;
	int l;
	int t;
	HttpInfo *td;

	td = (HttpInfo *)(uf->tdata);
	l = strlen(uf->parse->url)+40;
	if (l>gbuflen) {
		if (gbuf)
			UrlFree(gbuf);
		gbuf = (char *)UrlMalloc(l);
		if (!gbuf)
			return UrlSetErrno(uf->ctx,ENOMEM);
		gbuflen = l;
	}
	p = gbuf;
	strcpy(p,cmd);
	p += strlen(p);
	*p++ = ' ';
	if (uf->ctx->use_proxy) {
		char *s;
		s = UrlUnparseC(uf->parse);
		if (!s || !*s)
			s = "/";
		strcpy(p,s);
		p += strlen(p);
	} else {
		char *s = uf->parse->file;
		if (!s || !*s)
			s = "/";
		strcpy(p,s);
		p += strlen(p);
		if (uf->parse->query) {
			*p++ = '?';
			strcpy(p,uf->parse->query);
			p += strlen(p);
		}
	}
	if (postfix && postfix[0]) {
		*p++ = ' ';
		strcpy(p,postfix);
		p += strlen(p);
	}
	strcpy(p,crlf);
	p += strlen(p);
	if (addheader) {
		/* Add the rest of the header plus the blank line */
		/* TBD - other header stuff? */
		/* TBD - send "Referer: <url of source doc or fwd ref>" */
		/* TBD - send "From: user@host" ? */
		strcpy(p,crlf);		/* blank line means end of header */
		p += strlen(p);
	}
	UrlMessage(uf->ctx,URL_MSG_CONNECT,"Sending HTTP request");
	if (uf->ctx->debug)
		UrlMessagef(uf->ctx,URL_MSG_DEBUG,"Send to HTTP: %s",gbuf);
	t = write(uf->fd,gbuf,p-gbuf);
	if (t<0)
		return UrlSetErrno(uf->ctx,errno);
	return 0;	/* header successfully sent */
}

int
UrlHttpClose(uf)
UFILE *uf;
{
	int t;
	HttpInfo *td;

	td = (HttpInfo *)(uf->tdata);
	if (td->st) {
		UrlSymTabFree(uf->ctx,td->st);
		td->st = (UrlSymTab *)0;
	}
	if (uf->fd>=0) {
		t = close(uf->fd);
		uf->fd = -1;
	} else
		t = 0;
	return t;
}

char *
UrlHttpGetMeta(uf,name)
UFILE *uf;
char *name;
{
	HttpInfo *td;
	UrlSymTab *st;
	char *v;

	td = (HttpInfo *)(uf->tdata);
	st = td->st;
	if (!st) {
		UrlSetErrno(uf->ctx,EINVAL);
		return (char *)0;
	}
	v = UrlSymFind(uf->ctx,st,name);
	if (!v)
		UrlSetErrno(uf->ctx,EINVAL);
	return v;
}

/* end */
