/* Copyright 1995-1996 Jim McBeath
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that my name not be used in advertising
 * or publicity pertaining to distribution of the software without
 * specific, written prior permission.  I make no representations about
 * the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * I disclaim all warranties with regard to this software, including all
 * implied warranties of merchantability and fitness.  In no event shall
 * I be liable for any special, indirect or consequential damages or any
 * damages whatsoever resulting from loss of use, data or profits, whether
 * in an action of contract, negligence or other tortious action, arising
 * out of or in connection with the use or performance of this software.
 */
/* urllfile.c - localfile access
 *
 * Jim McBeath, October 14, 1994
 */

#include "urlopen.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <dirent.h>
#include <fcntl.h>	/* O_* flags */

extern char *strrchr();

/* forward decls */
int UrlLocalfileOpen();
int UrlLocalfileClose();
char *UrlLocalfileGetLine();
char *UrlLocalfileGetMeta();
int UrlLocalfileIsLocalFile();
int UrlLocalfileIsFiltered();

typedef struct localfileinfo {
	int isdir;
	int state;
	DIR *dir;
	long size;
} LocalfileInfo;

void
UrlLocalfileRegister(ctx)
UrlContext *ctx;
{
	UACCESS *access;

	access = UrlRegisterNew(ctx,"localfile");
	if (!access) return;
	access->datasize = sizeof(LocalfileInfo);
	access->open = UrlLocalfileOpen;
	access->close = UrlLocalfileClose;
	access->getline = UrlLocalfileGetLine;
	access->getmeta = UrlLocalfileGetMeta;
	access->islocalfile = UrlLocalfileIsLocalFile;
	access->isfiltered = UrlLocalfileIsFiltered;
}

int		/* 0 if OK */
UrlLocalfileOpen(uf)
UFILE *uf;
{
	struct stat sb;
	int t;
	LocalfileInfo *td;

	td = (LocalfileInfo *)(uf->tdata);
	t = stat(uf->parse->file,&sb);
	if (t)
		return UrlSetErrno(uf->ctx,errno); /* stat failed */
	if S_ISDIR(sb.st_mode) {
		td->isdir = 1;
		UrlParseMakeTrailingSlash(uf->parse);
		td->dir = opendir(uf->parse->file);
		if (!td->dir) {
			return UrlSetErrno(uf->ctx,errno);
		}
	} else {	/* TBD - check for symlink? */
		/* plain file */
		int flags;
		if (uf->mode[0]=='w')
			flags = O_WRONLY;
		else
			flags = O_RDONLY;
		uf->fd = open(uf->parse->file,flags,0);
		if (uf->fd<0) {
			return UrlSetErrno(uf->ctx,errno);
		}
		td->size = sb.st_size;
	}
	return 0;
}

int
UrlLocalfileClose(uf)
UFILE *uf;
{
	int t;
	LocalfileInfo *td;

	td = (LocalfileInfo *)(uf->tdata);
	if (uf->fd>=0) {
		t = close(uf->fd);
		uf->fd = -1;
	} else if (td->dir) {
		td->state = -1;
		t = closedir(td->dir);
		td->dir = (DIR *)0;
	} else
		t = 0;
	return t;
}

int
UrlLocalfileIsLocalFile(uf)
UFILE *uf;
{
	LocalfileInfo *td;

	td = (LocalfileInfo *)(uf->tdata);
	if (uf->fd>=0 || td->dir)
		return 1;
	return 0;	/* no file */
}

int
UrlLocalfileIsFiltered(uf)
UFILE *uf;
{
	LocalfileInfo *td;

	td = (LocalfileInfo *)(uf->tdata);
	return td->isdir;
}

#ifndef S_ISLNK
#define S_ISLNK(mode) (((mode)&S_IFMT) == S_IFLNK)
#endif

char *
UrlLocalfileDirFileHtml(uf)
UFILE *uf;
{
	UrlDirFileInfo fileinfo;
	struct dirent *de;
	char *s;
	struct stat sb;
#ifndef MAXPATHLEN
#define MAXPATHLEN 1024
#endif
	char fname[MAXPATHLEN+1];
	char abuf[MAXPATHLEN+1];
	char *ds;
	char datebuf[40];
	LocalfileInfo *td;
	int t;
	time_t currenttime;

	td = (LocalfileInfo *)(uf->tdata);
	time(&currenttime);
	while ((de=readdir(td->dir))!=0) {
		if (strcmp(de->d_name,".")==0)
			continue;	/* skip "." */
		sprintf(fname,"%s/%s",uf->parse->file,de->d_name);
		t = lstat(fname,&sb);	/* we want to see S_ISLNK if symlink */
		if (t) continue;	/* ignore files we can't stat */
		memset((char *)(&fileinfo),0,sizeof(UrlDirFileInfo));
		fileinfo.name = de->d_name;
		if (S_ISLNK(sb.st_mode)) {
			abuf[0];
			(void)readlink(fname,abuf,sizeof(abuf));
			if (abuf[0])
				fileinfo.symlink = abuf;
		} else {
			/* not a symbolic link */
			fileinfo.symlink = (char *)0;
		}
		ds = ctime(&(sb.st_mtime));
		if (currenttime-sb.st_mtime>15811200) {
			/* over about 6 months old */
			sprintf(datebuf,"%6.6s  %4.4s", ds+4, ds+20);
		} else {
			/* less than about 6 months old */
			sprintf(datebuf,"%6.6s %5.5s", ds+4, ds+11);
		}
		if S_ISDIR(sb.st_mode)
			fileinfo.isdir = 1;
		if (sb.st_mode & (S_IXUSR|S_IXGRP|S_IXOTH))
			fileinfo.isexec = 1;
		fileinfo.date = datebuf;
		fileinfo.size = sb.st_size;
		s = UrlDirFileHtml(uf,&fileinfo);
		if (s)
			return s;
	}
	return (char *)0;
}

char *
UrlLocalfileGetLine(uf)
UFILE *uf;
{
	LocalfileInfo *td;
#define LOCALFILEHTMLBUFSIZE 1200
	static char buf[LOCALFILEHTMLBUFSIZE];
	char *s;

	td = (LocalfileInfo *)(uf->tdata);
	if (!td->isdir) {
		if (!uf->fd) {
			return (char *)0;
		}
		return UrlUGetLine(uf);
	}
	switch (td->state) {
	case 0:	/* output header */
		td->state = 10;		/* continue header on next call */
		sprintf(buf,"<HTML><HEADER><TITLE>Directory Listing for %s</TITLE></HEADER>\n",
			uf->parse->file);
		return buf;
	case 10:
		td->state++;
		sprintf(buf,"<BODY><H1>Directory Listing for %s</H1>\n",
			uf->parse->file);
		return buf;
	case 11:
		td->state = 1;	/* ready to do files */
		return "<pre>\n";
	case 1:	/* do files */
		/* TBD - sort directory? */
		s = UrlLocalfileDirFileHtml(uf);
		if (s)
			return s;
		/* end of directory, clean up */
		td->state = -1;
		closedir(td->dir);
		td->dir = (DIR *)0;
		return "</pre>\n";
	default:
		errno = 0;
		return (char *)0;
	}
}

char *
UrlLocalfileGetMeta(uf,name)
UFILE *uf;
char *name;
{
	UrlContext *ctx;
	LocalfileInfo *td;

	td = (LocalfileInfo *)(uf->tdata);
	ctx = uf->ctx;
	if (strcmp(name,"content-type")==0) {
		UPARSE *parse;
		char *ext, *ct;

		if (td->isdir)
			return "text/html";  /* directories filtered to HTML */
		parse = uf->parse;
		if (!parse->file)
			return "";	/* no type specified */
		ext = strrchr(parse->file,'.');
		if (!ext)
			return "";
		ct = UrlGetExtContType(ctx,ext+1);
		return ct;
	}
	UrlSetErrno(ctx,EINVAL);
	return (char *)0;
}

/* end */
