/* Copyright 1995 Jim McBeath
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that my name not be used in advertising
 * or publicity pertaining to distribution of the software without
 * specific, written prior permission.  I make no representations about
 * the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * I disclaim all warranties with regard to this software, including all
 * implied warranties of merchantability and fitness.  In no event shall
 * I be liable for any special, indirect or consequential damages or any
 * damages whatsoever resulting from loss of use, data or profits, whether
 * in an action of contract, negligence or other tortious action, arising
 * out of or in connection with the use or performance of this software.
 */
/* urlopen.c - open a url based on access type
 *
 * Jim McBeath, October 14, 1994
 */

#include "urlopen.h"

extern char *strchr();
extern char *getenv();

extern char *UrlFGetLine();

/* forward decls */
void UrlFreeUfile();

UFILE *
UrlNewUfile(access)
UACCESS *access;
{
	UrlContext *ctx;
	UFILE *uf;

	ctx = access->ctx;
	if (!ctx->ufilefreelist) {
		/* TBD - make this allocate more than one at a time? */
		ctx->ufilefreelist = (UFILE *)UrlMalloc(sizeof(UFILE));
		if (!ctx->ufilefreelist) {
			UrlSetErrno(ctx,ENOMEM);
			return (UFILE *)0;
		}
		memset(ctx->ufilefreelist,0,sizeof(UFILE));
	}
	uf = ctx->ufilefreelist;
	ctx->ufilefreelist = uf->next;
	uf->next = 0;
	uf->access = access;
	uf->fd = -1;
	if (access->datasize) {
		void *p;
		p = UrlMalloc(access->datasize);
		if (!p) {
			UrlSetErrno(ctx,ENOMEM);
			UrlFreeUfile(uf);
			return (UFILE *)0;
		}
		memset(p,0,access->datasize);
		uf->tdata = p;
	}
	return uf;
}

void
UrlFreeUfile(uf)
UFILE *uf;
{
	UrlContext *ctx;

	ctx = uf->ctx;
	if (uf->access->datasize && uf->tdata) {
		if (uf->access->tdatafree)
			(*(uf->access->tdatafree))(uf);
		UrlFree(uf->tdata);
	}
	if (uf->ibuf)
		UrlFree(uf->ibuf);
	if (uf->parse)
		UrlFreeParse(uf->parse);
	memset(uf,0,sizeof(UFILE));
	uf->next = ctx->ufilefreelist;
	ctx->ufilefreelist = uf;
}

/* Opens an absolute (fully qualified) URL */
UFILE *
UrlOpen(ctx,url,mode)
UrlContext *ctx;
char *url;
char *mode;
{
	char *c;
	UACCESS *ut;
	UFILE *uf;
	UPARSE *uparse;
	int t;

	if (!url || !*url) {
		UrlSetErrno(ctx,EINVAL);
		return (UFILE *)0;
	}
	uparse = UrlParse(ctx,url);
	if (!uparse)
		return (UFILE *)0;
	if (ctx->use_proxy)
		ut = UrlFindAccess(ctx,ctx->use_proxy->access);
	else
		ut = UrlFindAccess(ctx,uparse->access);
	if (!ut)
		return (UFILE *)0;
	uf = UrlNewUfile(ut);
	if (!uf)
		return (UFILE *)0;
	uf->ctx = ctx;
	uf->parse = uparse;
	uf->mode = UrlStrdup(mode);
	if (!uf->mode) {
		UrlSetErrno(ctx,ENOMEM);
		UrlFreeUfile(uf);
		return (UFILE *)0;
	}
	t = (*ut->open)(uf);		/* access-specific open */
	if (t) {
		UrlFreeUfile(uf);	/* open failed */
		return (UFILE *)0;
	}
	return uf;
}

int
UrlClose(uf)
UFILE *uf;
{
	int t;

	if (uf->access->close) {
		t = (*uf->access->close)(uf);
	} else {
		if (uf->fd>=0) {
			t = close(uf->fd);
			uf->fd = -1;
		} else {
			t = 0;
		}
	}
	UrlFreeUfile(uf);
	return t;
}

FILE *
UrlGetFile(uf)
UFILE *uf;
{
	FILE *f;

	if (uf->fd<0) {
		UrlSetErrno(uf->ctx,EINVAL);
		return (FILE *)0;
	}
	f = fdopen(uf->fd,uf->mode);
	if (!f)
		return f;	/* failed */
	if (uf->ibuf && (uf->ibuf_get<uf->ibuf_end)) {
		/* set up the buffer with leftover data from our stream */
		setvbuf(f,uf->ibuf,_IOFBF,uf->ibuf_alloc);
/* OS-dependent */
#ifdef linux
#define PTR _IO_read_ptr
#define END _IO_read_end
		f->PTR = uf->ibuf_get;
		f->END = uf->ibuf_end;
#else
#if defined(hpux) || defined(__H3500) || defined(__H3050R)
#define CNT __cnt
#define PTR __ptr
#else
#define CNT _cnt
#define PTR _ptr
#endif
		f->PTR = uf->ibuf_get;
		f->CNT = uf->ibuf_end - uf->ibuf_get;
#endif
	}
	return f;
}

int
UrlIsFiltered(uf)
UFILE *uf;
{
	int t;

	if (!uf->access->isfiltered)
		return 0;	/* no method, must not be filtered */
	t = (*uf->access->isfiltered)(uf);
	return t;
}

char *
UrlGetLine(uf)
UFILE *uf;
{
	char *s;

	if (!uf->access->getline) {
		FILE *f;
		f = UrlGetFile(uf);
		if (!f)
			return (char *)0;
		s = UrlFGetLine(f);
		uf->bytesread += strlen(s);
		UrlShowBytesRead(uf);
	} else {
		s = (*uf->access->getline)(uf);
	}
	return s;
}

int
UrlIsLocalFile(uf)
UFILE *uf;
{
	int t;

	if (!uf->access->islocalfile)
		return 0;	/* no method, must not be local */
	t = (*uf->access->islocalfile)(uf);
	return t;
}

FILE *
UrlGetLocalFile(uf)
UFILE *uf;
{
	FILE *f, *of;
	static char *tmpdir=0;
	char oname[1024];
	int c;

	if (UrlIsLocalFile(uf)) {
		f = UrlGetFile(uf);
		return f;	/* it's already a local file */
	}

	/* copy to local file, open local file, return pointer */
	if (!tmpdir) {
		/* TBD - different way to specify tmp dir or mktmp pattern? */
		tmpdir = getenv("TMPDIR");
		if (!tmpdir)
			tmpdir = "/tmp";
	}
	strcpy(oname,tmpdir);
	strcat(oname,"/urlXXXXXX");
	mktemp(oname);
	of = fopen(oname,"w+");
	if (!of)
		return (FILE *)0;	/* can't get it */
	while (!UrlUeof(uf)) {
		c = UrlUgetc(uf);
		if (c==EOF) break;
		fputc(c,of);
	}
	rewind(of);
#if 0	/* TBD - will this delete file later as desired? */
	unlink(oname);	/* deletes it after caller closes the file */
#endif
	return of;
}

/* end */
