/* Copyright 1995-1998 Jim McBeath
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that my name not be used in advertising
 * or publicity pertaining to distribution of the software without
 * specific, written prior permission.  I make no representations about
 * the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * I disclaim all warranties with regard to this software, including all
 * implied warranties of merchantability and fitness.  In no event shall
 * I be liable for any special, indirect or consequential damages or any
 * damages whatsoever resulting from loss of use, data or profits, whether
 * in an action of contract, negligence or other tortious action, arising
 * out of or in connection with the use or performance of this software.
 */
/* urlproxy.c - simple proxy server
 *
 * Jim McBeath, November 15, 1995
 */

#include "urlopen.h"
#include <signal.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef _AIX
#include <sys/select.h>
#endif
#include <fcntl.h>

#ifdef O_NONBLOCK
#define FNDELAY_OR_NONBLOCK O_NONBLOCK
#else
#define FNDELAY_OR_NONBLOCK FNDELAY
#endif

extern char *strchr();

void UrlProxyPrintMessage();
void UrlProxyProcessAccept();
void UrlProxyWriteHeader();
void UrlProxyWriteError();

typedef struct _urlproxyproginfo {
	char *progname;
	UrlContext *uctx;
	int verbose;
	int nofork;
	int redirect_limit;
	int port;
	int listensocket;
} UrlProxyProgInfo;

static int UrlProxyInterruptCount;
static int UrlProxySigpipeCount;

void
UrlProxyHandleInterrupt()
{
	UrlProxyInterruptCount++;
	signal(SIGINT,(void (*))0);
}

void
UrlProxyHandleSigpipe()
{
	UrlProxySigpipeCount++;
	signal(SIGPIPE,UrlProxyHandleSigpipe);
}

/* Don't fear the reaper */
void
UrlProxyReap(proginfo)
UrlProxyProgInfo *proginfo;
{
	int pid;
	int status;

/* old definition using wait3:
#define WAITANY(statusp,options) wait3(statusp,options,NULL)
*/
#define WAITANY(statusp,options) waitpid(-1,statusp,options)
	while ((pid=WAITANY(&status, WNOHANG))>0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Process %d reaped\n",pid);
	}
}

void
UrlProxyCleanup(proginfo)
UrlProxyProgInfo *proginfo;
{
	shutdown(proginfo->listensocket,2);
	close(proginfo->listensocket);
}

void
UrlProxySetNonBlock(
	int fd,
	UrlProxyProgInfo *proginfo)
{
	int t;

	t = fcntl(fd, F_SETFL, FNDELAY_OR_NONBLOCK|fcntl(fd,F_GETFL,0));
		/* set non-blocking */
	if (t<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Non-block fcntl errno=%d",errno);
	}
}

void
UrlProxySetBlock(
	int fd,
	UrlProxyProgInfo *proginfo)
{
	int t;

	t = fcntl(fd, F_SETFL, (~FNDELAY_OR_NONBLOCK)&fcntl(fd,F_GETFL,0));
	if (t<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Re-block fcntl errno=%d",errno);
	}
}

int
UrlProxy(argc,argv)
int argc;
char *argv[];
{
	UrlProxyProgInfo proginfo;
	int i;
	int s, t;

	memset((char *)(&proginfo),'\0',sizeof(proginfo));
	proginfo.progname = argv[0];
	t = UrlInit(&proginfo.uctx);/* init our package with standard methods */
/* TBD - may want to omit localfile access method for security */
	if (t) {
		return t;
	}
#define URL_REDIRECT_LIMIT 100
	proginfo.redirect_limit = URL_REDIRECT_LIMIT;
	UrlInitSigPipe();
	signal(SIGINT,UrlProxyHandleInterrupt);
	signal(SIGPIPE,UrlProxyHandleSigpipe);
	proginfo.uctx->appdata = (void *)(&proginfo);
	proginfo.uctx->is_proxy = 1;
	UrlSetMessageFunc(proginfo.uctx,UrlProxyPrintMessage);
	if (argc<=1)
		return UrlProxyUsage(&proginfo);
	for (i=1; i<argc; i++) {
		if (argv[i][0]=='-') {
			if (strcmp(argv[i],"-debug")==0)
				proginfo.uctx->debug = 1;
			else if (strcmp(argv[i],"-cut")==0) {
				extern void *malloc();
				int j;
				char **oldargv = argv;
				/* copy argv into another array */
				char **newargv= malloc(sizeof(char *)*(argc+1));
				for (j=0; j<=argc; i++)
					newargv[j] = argv[j];
				argv = newargv;	/* point to our copy */
				/* now clear out the args in the original */
				for (j=i; j<argc; j++)
					oldargv[j] = " ";
						/* hide remaining args */
			}
			else if (strcmp(argv[i],"-icon_prefix")==0) {
				if (++i>=argc)
					return UrlCatUsage(&proginfo);
					    /* missing arg for -icon_prefix */
				UrlSetIconPrefix(proginfo.uctx,argv[i]);
			}
			else if (strcmp(argv[i],"-nofork")==0)
				proginfo.nofork = 1;
			else if (strcmp(argv[i],"-passive")==0)
				proginfo.uctx->passive = 1;
			else if (strcmp(argv[i],"-port")==0) {
				int n = atoi(argv[++i]);
				if (n<=0)
					return UrlProxyUsage(&proginfo);
				proginfo.port = n;
			}
			else if (strcmp(argv[i],"-proxy")==0) {
				int l; char *s;
				if (++i>=argc)
					return UrlProxyUsage(&proginfo);
						/* missing arg for -proxy */
				l = strlen(argv[i])+30;
				s = UrlMalloc(l);
				strcpy(s,"http://");
				if (!strchr(argv[i],':'))
					strcat(s,"localhost:");/* default host*/
				strcat(s,argv[i]); /* port number */
				UrlSetProxy(proginfo.uctx,s);
				UrlFree(s);
			}
#ifdef TERM
			else if (strcmp(argv[i],"-term")==0)
				proginfo.uctx->useterm = 1;
#endif
			else if (strcmp(argv[i],"-verbose")==0)
				proginfo.verbose = 1;
			else if (strcmp(argv[i],"-version")==0)
				return UrlProxyVersion(&proginfo);
			else
				return UrlProxyUsage(&proginfo);
		} else {
			return UrlProxyUsage(&proginfo);
		}
	}
	if (!proginfo.port) {
		/* port number is required */
		return UrlProxyUsage(&proginfo);
	}
	t = UrlProxyWork(&proginfo);
	return t;	/* only returns here on error */
}

int
UrlProxyWork(proginfo)
UrlProxyProgInfo *proginfo;
{
	int s;
	int portnum = proginfo->port;
	UrlContext *ctx = proginfo->uctx;

	s = UrlListenPort(ctx,portnum);
	if (s<0) {
		fprintf(stderr,"Can't listen: %s\n", UrlGetErrorString(ctx));
		return 1;
	}
	UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
		"Listening on port=%d, fd=%d",portnum,s);
	proginfo->listensocket = s;
	while (UrlProxyInterruptCount==0) {
		UrlProxyProcessAccept(proginfo);
	}
	UrlProxyCleanup(proginfo);
	return 0;
}

void
UrlProxyProcessAccept(proginfo)
UrlProxyProgInfo *proginfo;
{
	int sk;
	int pid;
	int t;

	UrlMessage(proginfo->uctx,URL_MSG_DEBUG,"Waiting for accept");
	sk = UrlAcceptSocket(proginfo->uctx,proginfo->listensocket);
	if (sk<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Accept failed, errno=%d",errno);
		return;	/* ignore error connections */
	}
	UrlProxyReap(proginfo);	/* clean up any exited children */
	UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,"Accept on fd=%d",sk);
	if (proginfo->nofork)
		pid = 0;	/* for debugging - don't fork */
	else
		pid = fork();	/* child process to handle each request */
	if (pid<0) {
		/* error */
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Error on fork: %d",errno);
		return;
	} else if (pid>0) {
		/* in parent process */
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Started child process pid=%d",pid);
		close(sk);	/* only used in the child */
		return;		/* back to main wait loop */
	}
	/* in child process (or when -nofork specified) */
	UrlMessage(proginfo->uctx,URL_MSG_DEBUG,"Child is running");
	t = UrlProxyProcessRequest(proginfo,sk);
	UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,"Child exit status=%d",t);
	if (proginfo->nofork)
		return;	/* allow another connection */
	exit(t);	/* exit from child */
}

int
UrlProxyProcessRequest(proginfo,sk)
UrlProxyProgInfo *proginfo;
int sk;
{
	char *s;
	char *cmd;
	char *line;
	char *url;
	FILE *sr, *sw;
	UFILE *uf;
	int t;
	int httpver;

	UrlProxySigpipeCount = 0;
	sr = fdopen(sk,"r");
	if (!sr) {
		UrlMessage(proginfo->uctx,URL_MSG_DEBUG,
			"Can't fdopen socket for read");
		return 1;
	}
	sw = fdopen(dup(sk),"w");
		/* use a different file descriptor so that we can set sk
		 * to non-blocking for the read, but still block on the
		 * write. */
	if (!sw) {
		UrlMessage(proginfo->uctx,URL_MSG_DEBUG,
			"Can't fdopen socket for write");
		return 1;
	}

	httpver = UrlProxyReadCommand(proginfo,sr,&cmd,&url);
	if (httpver<0)
		return -1;	/* error reading command */
#if 0 /* whatever the request, we'll forward it once we make the connection */
	if ((strcmp(cmd,"GET")!=0) && (strcmp(cmd,"HEAD")!=0)) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Unknown command '%s' from client",cmd);
		return 1;
	}
#endif

	/* Call UrlOpen to get requested data */
	uf = UrlOpen(proginfo->uctx,url,"r");
	if (!uf) {
		/* output error header to client */
		char *e = UrlGetErrorString(proginfo->uctx);
		UrlProxyWriteError(sw,e);
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Error opening url '%s' from client",url);
		fflush(sw);
		fclose(sw);
		fclose(sr);
		return 1;
	}

	if (strcmp(uf->parse->access,"http")==0) {
	/* If it's an HTTP url, we need to forward all of our extra
	 * input from the client to the server, and returns its
	 * header lines; so we just make the connection, and
	 * do the transfer raw in both directions. */
		char postfix[20];
		if (httpver)
			sprintf(postfix,"HTTP/%d.0",httpver);
		else
			postfix[0] = 0;
		t = UrlHttpSendCmd(uf,cmd,postfix,0);
		if (t<0) {
			UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
				"Error passing request to http server: %s",
				UrlGetErrorString(proginfo->uctx));
			return 1;
		}
		if (httpver>0) {	/* transfer rest of header */
		    while (((line=UrlFGetLine(sr))!=0)&&*line) {
			t = write(uf->fd,line,strlen(line));
			if (t<0) {
			    UrlSetErrno(proginfo->uctx,errno);
			    UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
				"Error passing header to http server: %s",
				UrlGetErrorString(proginfo->uctx));
			    return 1;
			}
			if (line[0]=='\n'||line[0]=='\r')
				break;	/* stop after sending blank line */
		    }
		}
		/* Now we assume that what we get back from the server will
		 * include the header info, so we don't output it separately. */
	} else {
		UrlProxyWriteHeader(sw,uf,httpver);
	}

	/* Read whatever comes back from the connection and deliver to client */
	UrlProxyTransferData(proginfo,sr,sw,uf);

	UrlMessage(proginfo->uctx,URL_MSG_DEBUG,
		"Ready to call UrlClose(uf)\n");

	t = UrlClose(uf);
	if (t) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"Error calling UrlClose: %s",
			UrlGetErrorString(proginfo->uctx));
		return 1;
	}

	UrlMessage(proginfo->uctx,URL_MSG_DEBUG,
		"Ready to flush and close in child\n");

	/* finish up child */
	t = fflush(sw);
	if (t<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"fflush(sw) errno=%d",errno);
	}

#if 0
	t = shutdown(sk,0);
	if (t<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"shutdown(sk,0) errno=%d",errno);
	}
	t = close(sk);
	if (t<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"close(sk) errno=%d",errno);
	}
#endif

	t = fclose(sw);
	if (t<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"fclose(sw) errno=%d",errno);
	}
	t = fclose(sr);
	if (t<0) {
		UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
			"fclose(sr) errno=%d",errno);
	}
	return 0;
}

/* Transfer data from source to client */
UrlProxyTransferData(proginfo,sr,sw,uf)
UrlProxyProgInfo *proginfo;
FILE *sr;
FILE *sw;
UFILE *uf;
{
#define TRANBUFSIZE 1000
	char buf[TRANBUFSIZE];
	char *s;
	int c;
	int n;
	int nn;
	int cfd;	/* client-side read file descriptor */
	int sfd;	/* server-side read file descriptor */
	int nfds;
	fd_set rfds;
	int bytesToClient=0;
	int t;

	cfd = fileno(sr);
	sfd = UrlUfileno(uf);

	UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,"Tansferring data\n");

	/* We come in with a FILE* for the client, which may have buffered
	 * data in it.  Rather than poking inside to find out if there's
	 * any data there, we just use fread until it says "no more",
	 * assuming that won't be a problem since we have non-blocking set
	 * on that descriptor. */
	do {
		UrlProxySetNonBlock(fileno(sr),proginfo);
		n = fread(buf,1,sizeof(buf)-2,sr);
		UrlProxySetBlock(fileno(sr),proginfo);
		if (n>0)
			write(sfd,buf,n);
	} while (n>0);

	nfds = (cfd>sfd?cfd:sfd)+1;
	while (!UrlUeof(uf)) {
		FD_ZERO(&rfds);
		if (!feof(sr))
			FD_SET(cfd,&rfds);
		FD_SET(sfd,&rfds);
		/* wait for one side or the other to have something for us */
		n = select(nfds,&rfds,(fd_set*)0,(fd_set*)0,(struct timeval*)0);
		if (n<0) {
			if (errno==EINTR) {
				continue;	/* ignore interrupt */
			}
			UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
				"Select errno=%d",errno);
			continue;
		}
		if (FD_ISSET(cfd,&rfds)) {
			/* read from client, send to server */
			UrlProxySetNonBlock(cfd,proginfo);
			n = read(cfd,buf,sizeof(buf)-2);
			UrlProxySetBlock(cfd,proginfo);
			if (n==0) {
				/* EOF on client side, user probably pushed
				 * the Stop button in the browser.
				 * Since client is no longer talking to us,
				 * there's nothing left for us to transfer.
				 */
				return;
			}
			if (n>0)
				write(sfd,buf,n);
			if (UrlProxySigpipeCount>0)
				return;	/* server disconnected */
		}
		if (FD_ISSET(sfd,&rfds)) {
			/* read from server, send to client */
			UrlUFillBuf(uf);
		}
		if (UrlIsFiltered(uf)) {
			/* This will block, but it's too hard right now to
			 * make it non-blocking */
			s = UrlGetLine(uf);
			while (s) {
				fputs(s,sw);
				bytesToClient += strlen(s);
				s = UrlGetLine(uf);
			}
		} else {
			n = UrlUBufferCount(uf);
			while (n>0) {
				if (n>sizeof(buf)-2)
					n = sizeof(buf)-2;
				n = UrlUread(buf,1,n,uf);
				if (n>0) {
					buf[n] = 0;
					nn = fwrite(buf,1,n,sw);
					if (nn!=n) {
						UrlMessagef(
						   proginfo->uctx,URL_MSG_DEBUG,
"Write error to client: tried %d bytes, wrote %d bytes\n",n,nn);
					}
					bytesToClient += nn;
					t = fflush(sw);	/* keep client busy */
					if (t!=0) {
						UrlMessagef(
						   proginfo->uctx,URL_MSG_DEBUG,
						   "fflush errno=%d\n",errno);
					}
					if (UrlProxySigpipeCount>0)
						return; /* client disconn */
				}
				n = UrlUBufferCount(uf);
			}
		}
	}

#if 0 /* old way */
	if (UrlIsFiltered(uf)) {
		char *s;
		s = UrlGetLine(uf);
		while (s) {
			fputs(s,sw);
			s = UrlGetLine(uf);
		}
	} else {
		while (!UrlUeof(uf) && ferror(sw)==0) {
			n = UrlUBufferCount(uf); /* chars in read buffer */
			n = TRANBUFSIZE;
			n = UrlUread(buf,1,n,uf);
			if (n>0) {
				buf[n] = 0;
				fwrite(buf,1,n,sw);
				fflush(sw);	/* keep client busy */
			}
		}
	}
#endif
	UrlMessagef( proginfo->uctx,URL_MSG_DEBUG,
		"Wrote %d bytes to client\n", bytesToClient);
}

int	/* http version (0 or 1) of requestor, -1 on error */
UrlProxyReadCommand(proginfo,sr,cmdp,urlp)
UrlProxyProgInfo *proginfo;
FILE *sr;
char **cmdp;	/* RETURN command string (e.g. GET) */
char **urlp;	/* RETURN url */
{
	char *line,*line1;
	char *url, *cmd;
	char *p;
	int httpver;

	/* Read command from client */
	UrlMessage(proginfo->uctx,URL_MSG_DEBUG,
		"Child waiting for input from client");
	line = UrlFGetLine(sr);
	if (!line || !*line) {
		UrlMessage(proginfo->uctx,URL_MSG_DEBUG,"No input from client");
		return -1;
	}
	line1 = UrlStrdup(line);
	UrlMessagef(proginfo->uctx,URL_MSG_DEBUG,
		"Input from client: %s",line1);

	cmd = line1;
	p = strchr(line1,' ');
	if (!p) {
		UrlMessage(proginfo->uctx,URL_MSG_DEBUG,
			"No keyword in input from client");
		return -1;
	}
	*p++ = 0;
	url = p;
	p = strchr(url,' ');
	if (p) {
		*p++ = 0;	/* null terminate the url */
		/* p now points to the HTTP version string */
		if (strncmp(p,"HTTP/",5)==0)
			httpver = 1;
		else
			httpver = 0;	/* no headers returned for pre-v1 */
	} else {
		p = strchr(url,'\n');
		if (p) *p=0;	/* toss out newline at end of incoming url */
	}
	*urlp = url;
	*cmdp = cmd;
	return httpver;
}

void
UrlProxyWriteHeader(sw,uf,httpver)
FILE *sw;
UFILE *uf;
int httpver;
{
	char *s;

	if (httpver==0) {
		/* no header for really old clients */
		if (!s || strcmp(s,"text/plain")==0)
			fprintf(sw,"<PLAINTEXT>\n");
		return;
	}
	s = UrlGetMeta(uf,"firstline");
	if (s)
		fprintf(sw,"%s\n",s);
	else {
		int http_status = 200;	/* OK status */
		char *http_status_msg = "XXX";
		fprintf(sw,"HTTP/1.0 %d %s\n",http_status,http_status_msg);
	}
	s = UrlGetMeta(uf,"content-type");
	fprintf(sw,"Content-type: %s\n",s?s:"text/plain");
	/* TBD - need to forward the entire header from http servers */
	fprintf(sw,"\n");	/* end of header */
}

void
UrlProxyWriteError(sw,e)
FILE *sw;
char *e;
{
	int statusnum;

	statusnum = 400;
	fprintf(sw,"HTTP/1.0 %d %s\n",statusnum,e);
	fprintf(sw,"Content-type: text/plain\n");
	fprintf(sw,"\n");	/* end of header */
	fprintf(sw,"Error: %s\n",e);
}

void
UrlProxyPrintMessage(ctx,type,str)
UrlContext *ctx;
int type;
char *str;
{
	UrlProxyProgInfo *proginfo;

	proginfo = (UrlProxyProgInfo *)(ctx->appdata);
	if (ctx->debug&&type==URL_MSG_DEBUG ||
	    proginfo->verbose&&type!=URL_MSG_DEBUG)
		fprintf(stderr,"%s\n",str);
}

int
UrlProxyVersion(proginfo)
UrlProxyProgInfo *proginfo;
{
	printf("%s\n%s\n",UrlVersion(),UrlCopyright());
	return 0;
}

int
UrlProxyUsage(proginfo)
UrlProxyProgInfo *proginfo;
{
	fprintf(stderr,"usage: %s -port portnum [options] urllist\n",
		proginfo->progname);
	fprintf(stderr,"-debug     turn on debugging\n");
	fprintf(stderr,"-icon_prefix str     set prefix for internal icons\n");
	fprintf(stderr,"-nofork    no fork, exit after one request (debug)\n");
	fprintf(stderr,"-passive   use passive mode for FTP connections\n");
	fprintf(stderr,"-port      specify the port on which to listen\n");
#ifdef TERM
	fprintf(stderr,"-term      use term to make connections\n");
#endif
	fprintf(stderr,"-verbose   print progress messages to stderr\n");
	fprintf(stderr,"-version   print urlopen verion to stdout and exit\n");
	return 2;
}

#ifdef URL_PROXY_MAIN
main(argc,argv)
int argc;
char *argv[];
{
	int t;
	t = UrlProxy(argc,argv);
	exit(t);
}
#endif

/* end */
