/* Copyright 1995-1996 Jim McBeath
 *
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice appear in all copies and
 * that both that copyright notice and this permission notice appear in
 * supporting documentation, and that my name not be used in advertising
 * or publicity pertaining to distribution of the software without
 * specific, written prior permission.  I make no representations about
 * the suitability of this software for any purpose.  It is provided
 * "as is" without express or implied warranty.
 *
 * I disclaim all warranties with regard to this software, including all
 * implied warranties of merchantability and fitness.  In no event shall
 * I be liable for any special, indirect or consequential damages or any
 * damages whatsoever resulting from loss of use, data or profits, whether
 * in an action of contract, negligence or other tortious action, arising
 * out of or in connection with the use or performance of this software.
 */
/* urluread.c - parallels to FILE read routines, but using UFILE
 *
 * Jim McBeath, October 13, 1995
 */

#include "urlopen.h"
#include <errno.h>

extern char *strchr();

void
UrlSetPreReadFunc(ctx,func)
UrlContext *ctx;
int (*func)();
{
	ctx->prereadfunc = func;
}

int	/* true if EOF in input */
UrlUeof(uf)
UFILE *uf;
{
	return uf->eofflag;
}

int	/* return fileno suitable for select call */
UrlUfileno(uf)
UFILE *uf;
{
	return uf->fd;
}

int	/* number of chars in the buffer */
UrlUBufferCount(uf)
UFILE *uf;
{
	if (!uf->ibuf)
		return 0;
	return (uf->ibuf_end-uf->ibuf_get);
}

int	/* 1 if OK, 0 if EOF or error */
UrlUFillBuf(uf)
UFILE *uf;
{
	int n;

#define IBUFALLOCSIZE 5000
	if (!uf->ibuf) {
		uf->ibuf = UrlMalloc(IBUFALLOCSIZE);
		if (!uf->ibuf)
			return 0;
		uf->ibuf_alloc = IBUFALLOCSIZE;
	}
tryAgain:
	if (uf->ctx->prereadfunc) {
		n = (*(uf->ctx->prereadfunc))(uf);
		if (n==0) {	/* if cb func returns 0, we stop */
			UrlMessagef(uf->ctx,URL_MSG_DEBUG,
				"Prereadfunc returned 0");
			return 0;
		}
	}
	errno = 0;
	n = read(uf->fd, uf->ibuf, uf->ibuf_alloc-1);
	if (n<0) {
		/* TBD - handle error or interrupt */
		if (errno==EAGAIN || errno==EWOULDBLOCK)
			goto tryAgain;
		UrlMessagef(uf->ctx,URL_MSG_DEBUG,
			"Read failed, errno=%d",errno);
		uf->ibuf_get = uf->ibuf;
		uf->ibuf_end = uf->ibuf;
		*(uf->ibuf_end) = 0;
		return 0;
	}
	uf->ibuf_get = uf->ibuf;
	uf->ibuf_end = uf->ibuf+n;
	*(uf->ibuf_end) = 0;
	if (n==0) {
		/* could be EOF */
		if (errno==EAGAIN || errno==EWOULDBLOCK)
			goto tryAgain;
		if (errno)
			UrlMessagef(uf->ctx,URL_MSG_DEBUG,
				"Read returned 0 with errno=%d",errno);
		uf->eofflag = 1;
		UrlMessage(uf->ctx,URL_MSG_READING,"Document done");
		return 0;
	}
	uf->bytesread += n;
	UrlShowBytesRead(uf);
	return 1;	/* got something */
}

int	/* returns next char, or EOF */
UrlUgetc(uf)
UFILE *uf;
{
	if (uf->eofflag)
		return EOF;
	if (uf->ibuf_get>=uf->ibuf_end) {
		/* fill buffer and check for EOF */
		if (!UrlUFillBuf(uf))
			return EOF;
	}
	return (int)(*(uf->ibuf_get++));
}

/* Read until newline or until buf is full, with null at end */
char *		/* returns first arg */
UrlUgets(buf,size,uf)	/* like fgets */
char *buf;
int size;
UFILE *uf;
{
	unsigned char *p;
	int slen;
	char *bf=buf;
	int sz=size;

	if (size==0)
		return buf;
	buf[0] = 0;
	while (!UrlUeof(uf)) {
		if (uf->ibuf_get>=uf->ibuf_end) {
			if (!UrlUFillBuf(uf)) {
				UrlMessage(uf->ctx,URL_MSG_DEBUG,
					"No FillBuf in UrlUgets");
				break;	/* error */
			}
		}
		p = (unsigned char *)strchr(uf->ibuf_get,'\n');
		if (p)
			slen = p - uf->ibuf_get + 1;
		else
			slen = uf->ibuf_end - uf->ibuf_get;
		if (slen>sz-1)
			slen = sz-1;
		memcpy(bf,uf->ibuf_get,slen);
		uf->ibuf_get += slen;
		bf += slen;
		sz -= slen;
		if (sz==1 || p) break;
	}
	if (UrlUeof(uf))
		UrlMessage(uf->ctx,URL_MSG_DEBUG,"EOF in UrlUGets");
	if (bf==buf)
		UrlMessage(uf->ctx,URL_MSG_DEBUG,"UrlUGets returns empty str");
	*bf = 0;	/* null terminate */
	return buf;
}

int	/* number of items read */
UrlUread(buf,size,nitems,uf)	/* like fread */
char *buf;
int size;
int nitems;
UFILE *uf;
{
	int slen;
	char *bf=buf;
	int sz=size*nitems;

	if (size<0||nitems<0)
		return 0;
	while (!UrlUeof(uf) && sz>0) {
		if (uf->ibuf_get>=uf->ibuf_end) {
			if (!UrlUFillBuf(uf))
				break;	/* error or eof */
		}
		slen = uf->ibuf_end - uf->ibuf_get;
		if (slen>sz)
			slen = sz;
		memcpy(bf,uf->ibuf_get,slen);
		uf->ibuf_get += slen;
		bf += slen;
		sz -= slen;
	}
	return nitems-(sz/size);	/* number of items read */
}

/* Gets a line up to and including newline, no matter how long */
char *		/* belongs to us */
UrlUGetLine(uf)
UFILE *uf;
{
	static char *buf=0;
	static int buflen=0;
	int n;
#define CHUNKSIZE 1000
	if (!buf) {
		buflen = CHUNKSIZE;
		buf = UrlMalloc(buflen);
		if (!buf) {
			buflen = 0;
			UrlSetErrno(uf->ctx,ENOMEM);
			return 0;
		}
	}
	buf[0] = 0;
	UrlUgets(buf,buflen-1,uf);
	n = strlen(buf);
	if (n<buflen-1)
		return buf;
/* TBD - look for trailing backslash? */
	/* Well, we didn't get it in one read, need to try more */
/* TBD - do this later */
UrlMessage(uf->ctx,URL_MSG_DEBUG,"UrlUGetLine returning partial line");
return buf;
}

/* Like UrlUGetLine, but discards the newline (and CR before it) */
char *		/* belongs to us */
UrlUGetLineNoNl(uf)
UFILE *uf;
{
	char *l;
	int llen;

	l = UrlUGetLine(uf);
	uf->xccount = 0;
	if (!l || !*l)
		return l;
	llen = strlen(l);
	if (l[llen-1]=='\n') {
		l[--llen] = 0;
		uf->xccount++;
		if (llen>0 && l[llen-1]=='\015') {
			l[--llen] = 0;	/* throw out CR before LF */
			uf->xccount++;
		}
	}
	return l;
}

/* end */
